# My ultimate goal is to have a class with a set of methods (with
# fixed, pre-determined signatures), whose behaviour can be defined
# individually for each object of said class.
# ↓ (Illustrative example; obviously non-functional code.)

class C {
	method m(Int $i) { … }
}
my $c = C.new( m => method (Int $i) { $i × 2 } );

# Ideally, the signature given in the class declaration would need to
# be respected:

my $c = C.new( m => method (Block $b) { $b.() } );
#= ↑ Therefore, this should not be allowed.

# And preferably, this would work with multi-methods also, like so:

class D {
	multi method m(Int $i) { … }
	multi method m(Str $s) { … }
}
my $d = D.new( m => (method (Int $i) { $i × 2 },
                     method (Str $s) { $s x 2 }) );


#####################################################################
# Now for some examples with functional code…                       #
#####################################################################
# EXAMPLE 1: Attribute-based solution                               #
#####################################################################

# One way to imitate this behaviour is by putting code in attributes:

class C {
	has Routine $.m;
	method m(Int $i) { $!m.($i) }
}
my $c = C.new( m => sub (Int $i) { $i × 2 } );
say $c.m(12);
#=( ↑ This approach works, but some logic needs to be programmed
      manually. Namely checking if the routine stored at $!m truly
      has a matching signature. And multi-methods would need multiple
      attributes. Each with their own check. )


#####################################################################
# EXAMPLE 2: Role-based solution                                    #
#####################################################################

# Another way is by replacing classes and objects with roles and
# classes. This is more automated, but feels weird to me. Instead of
# a class, we provide a role first:

role R {
	method m(Int $i) { … }
}

# And then, instead of an object, we make a class which fills the
# role:

#| Doing R ensures that the method signatures match.
class C does R {
	method m(Int $i) { $i × 2 };
}
say C.m(12);

# At this point, we can use the method with ‘C.m()’.
# But in case our class makes use of attributes, we need to
# instantiate an object, since we can’t look up attributes in
# classes. (Class variables could be used instead, but I think those
# have different semantics.) Therefore:

my $c = C.new;
say $c.m(12);
#=( ↑ This feels weird, seeing as we create a whole class just to
      make a single object out of it. It works, but doesn’t feel
      conceptually right to me. It’s just a hack to have an object,
      so it shouldn’t need to be a whole class. )


#####################################################################
# EXAMPLE 3: Anothe role-based solution                             #
#####################################################################

# Roles can be mixed into objects, too! Unfortunately, in this case
# we can’t use them to set signature constraints, as existing
# signatures and their constraints are just overriden:

role R {
	method m(Int $i) { … }
}
class C does R {
	method m(Int $i) { $i × 2 }
}

my $c = C.new does role { method m(Str $s) { $s x 2 } };
say $c.m(‘haha’);
#=( ↑ This works. But for my purposes, it shouldn’t! )
